SystemOrganization addCategory: #SmaCC!
SystemOrganization addCategory: #'SmaCC-Runtime'!

Error subclass: #SmaCCParserError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SmaCC-Runtime'!

!SmaCCParserError commentStamp: '<historical>' prior: 0!
SmaCCParserException is the exception raised when a parsing error occurs. The description of the exception will be the error message and the parameter of the exception is the parser. With this information, you can insert a custom error message in your text view that you are parsing. For example, in VisualWorks, the following code will insert an error message into your text view:
	textController insertAndSelect: ex description , ' ->' at: ex parameter position!

Object subclass: #SmaCCParser
	instanceVariableNames: 'scanner currentToken errorToken stateStack nodeStack'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SmaCC-Runtime'!

!SmaCCParser commentStamp: '<historical>' prior: 0!
SmaCCParser is an abstract class that defines most of the parsing actions. Subclasses will define methods that specify their transitions and reduction actions. These are normally defined automatically when compiling the parser.

Subclasses must implement the following messages:
	accessing
		emptySymbolTokenId
		reduceTable
		transitionTable

Instance Variables:
	currentToken	<SmaCCToken>	the token last returned by the scanner that has not been shifted (reduce actions leave the current token alone)
	nodeStack	<OrderedCollection>	collection of items on stack. These items are specific to the parser and can be any object. 
	scanner	<SmaCCScanner>	our scanner
	stateStack	<OrderedCollection of: Integer>	the stack of states for our parser (standard LR state stack)

!

!SmaCCParser class methodsFor: 'private' stamp: ' 6/1/07 20:34'!
defaultStartingState
	^1! !

!SmaCCParser class methodsFor: 'instance creation' stamp: ' 6/1/07 20:34'!
on: aStream 
	| parser scanner |
	scanner := self scannerClass on: aStream.
	parser := self new.
	parser scanner: scanner.
	^parser! !

!SmaCCParser class methodsFor: 'accessing' stamp: ' 6/1/07 20:34'!
parse: aString 
	^self parse: aString startingAt: self defaultStartingState! !

!SmaCCParser class methodsFor: 'accessing' stamp: 'lr 1/6/2007 22:39'!
parse: aString onError: aBlock 
	^[self parse: aString] on: SmaCCParserError
		do: [:ex | ex return: (aBlock value: ex description value: ex tag position)]! !

!SmaCCParser class methodsFor: 'accessing' stamp: ' 6/1/07 20:34'!
parse: aString startingAt: anInteger 
	^self parseStream: (ReadStream on: aString) startingAt: anInteger! !

!SmaCCParser class methodsFor: 'accessing' stamp: 'lr 1/6/2007 22:39'!
parse: aString startingAt: anInteger onError: aBlock 
	^[self parse: aString startingAt: anInteger] on: SmaCCParserError
		do: [:ex | ex return: (aBlock value: ex description value: ex tag position)]! !

!SmaCCParser class methodsFor: 'accessing' stamp: ' 6/1/07 20:34'!
parseStream: aStream 
	^self parseStream: aStream startingAt: self defaultStartingState! !

!SmaCCParser class methodsFor: 'accessing' stamp: 'lr 1/6/2007 22:39'!
parseStream: aStream onError: aBlock 
	^[self parseStream: aStream] on: SmaCCParserError
		do: [:ex | ex return: (aBlock value: ex description value: ex tag position)]! !

!SmaCCParser class methodsFor: 'accessing' stamp: ' 6/1/07 20:34'!
parseStream: aStream startingAt: anInteger 
	| parser |
	parser := self on: aStream.
	parser setStartingState: anInteger.
	^parser parse! !

!SmaCCParser class methodsFor: 'accessing' stamp: 'lr 1/6/2007 22:39'!
parseStream: aStream startingAt: anInteger onError: aBlock 
	^[self parseStream: aStream startingAt: anInteger] 
		on: SmaCCParserError
		do: [:ex | ex return: (aBlock value: ex description value: ex tag position)]! !

!SmaCCParser class methodsFor: 'private' stamp: ' 6/1/07 20:34'!
scannerClass
	^self subclassResponsibility! !

!SmaCCParser methodsFor: 'private' stamp: ' 6/1/07 20:34'!
acceptAction
	^0! !

!SmaCCParser methodsFor: 'private' stamp: ' 6/1/07 20:34'!
actionFor: aSymbolIndex 
	^self actionForState: self currentState and: aSymbolIndex! !

!SmaCCParser methodsFor: 'private' stamp: ' 6/1/07 20:34'!
actionForCurrentToken
	^self actionFor: currentToken id first! !

!SmaCCParser methodsFor: 'private' stamp: 'lr 11/25/2009 18:47'!
actionForState: stateIndex and: aSymbolIndex
	| index row |
	row := self transitionTable at: stateIndex.
	^ row isArray 
		ifTrue: [ 
			(row at: 1) == 2 
				ifTrue: [
					index := self binarySearchInArray: row
						for: aSymbolIndex
						size: 1.
					index == 0 
						ifTrue: [ ^ self errorAction ]
						ifFalse: [ ^ row at: 2 ] ]
				ifFalse: [ 
					index := self binarySearchInArray: row
						for: aSymbolIndex
						size: 2.
					index == 0
						ifTrue: [ ^ self errorAction ]
						ifFalse: [ ^ row at: index - 1 ] ] ]
		ifFalse: [ 
			(row at: 1) == 0 
				ifTrue: [ 
					index := self binarySearchInBytes: row
						for: aSymbolIndex
						size: 2.
					index == 0 
						ifTrue: [ ^ self errorAction ] 
						ifFalse: [ ^ ((row at: 2) bitShift: 8) + (row at: 3) ] ]
				ifFalse: [ 
					index := self binarySearchInBytes: row
						for: aSymbolIndex
						size: 4.
					index == 0 
						ifTrue: [ ^ self errorAction ] 
						ifFalse: [ ^ ((row at: index - 2) bitShift: 8) + (row at: index - 1) ] ] ]! !

!SmaCCParser methodsFor: 'private' stamp: ' 6/1/07 20:34'!
actionMask
	^2r11! !

!SmaCCParser methodsFor: 'private' stamp: 'lr 11/25/2009 18:47'!
binarySearchInArray: aRow for: aSymbolIndex size: step 
	| start mid length midItem stop |
	start := 3.
	stop := aRow size.
	length := (stop - start) // step.
	[length > 4] whileTrue: 
			[length := length bitShift: -1.
			mid := length * step + start.
			midItem := aRow at: mid.
			midItem <= aSymbolIndex ifTrue: [start := mid] ifFalse: [stop := mid]].
	[start <= stop] whileTrue: 
			[(aRow at: start) == aSymbolIndex ifTrue: [^start].
			start := start + step].
	^0! !

!SmaCCParser methodsFor: 'private' stamp: 'lr 11/25/2009 18:48'!
binarySearchInBytes: aRow for: aSymbolIndex size: step
	| start mid length high low midItem stop |
	high := aSymbolIndex bitShift: -8.
	low := aSymbolIndex bitAnd: 16rFF.
	start := 4.
	stop := aRow size - 1.
	length := (stop - start) // step.
	[length > 4]
		whileTrue:
			[length := length bitShift: -1.
			mid := length * step + start.
			midItem := aRow at: mid.
			((midItem == high and: [(aRow at: mid + 1) <= low]) or: [midItem < high])
				ifTrue: [start := mid]
				ifFalse: [stop := mid]].
	[start <= stop]
		whileTrue:
			[((aRow at: start) == high and: [(aRow at: start + 1) == low]) ifTrue: [^start].
			start := start + step].
	^0! !

!SmaCCParser methodsFor: 'private-error handling' stamp: ' 6/1/07 20:34'!
checkForErrors
	"If we have an error correction installed, we might have handled the errors. If we did, we don't 
	want to return the result, so we raise a final exception that can't be proceeded."

	errorToken isNil ifTrue: [^self].
	currentToken := errorToken.
	self reportErrorMessage: 'Token not expected'! !

!SmaCCParser methodsFor: 'private' stamp: ' 6/1/07 20:34'!
currentState
	^stateStack last! !

!SmaCCParser methodsFor: 'private-error handling' stamp: ' 6/1/07 20:34'!
dismissErrorToken
	currentToken := nil.
	self getNextToken! !

!SmaCCParser methodsFor: 'private-error handling' stamp: ' 6/1/07 20:34'!
dismissStackTopForErrorRecovery
	stateStack removeLast.
	^nodeStack removeLast! !

!SmaCCParser methodsFor: 'accessing' stamp: ' 6/1/07 20:34'!
emptySymbolTokenId
	^scanner emptySymbolTokenId! !

!SmaCCParser methodsFor: 'private' stamp: ' 6/1/07 20:34'!
errorAction
	^3! !

!SmaCCParser methodsFor: 'private-error handling' stamp: ' 6/1/07 20:34'!
errorHandlerStates
	^stateStack collect: 
			[:each | 
			| action |
			action := self actionForState: each and: self errorTokenId.
			(action bitAnd: self actionMask) = 1 
				ifTrue: [action bitShift: -2]
				ifFalse: [0]]! !

!SmaCCParser methodsFor: 'accessing' stamp: ' 6/1/07 20:34'!
errorTable
	^#()! !

!SmaCCParser methodsFor: 'accessing' stamp: ' 6/1/07 20:34'!
errorTokenId
	^scanner errorTokenId! !

!SmaCCParser methodsFor: 'private' stamp: ' 6/1/07 20:34'!
findErrorHandlerIfNoneUseErrorNumber: anInteger 
	| handlerStates index startingErrorToken newStack |
	handlerStates := self errorHandlerStates reverse.
	startingErrorToken := currentToken.
	
	[index := (1 to: handlerStates size) detect: 
					[:each | 
					| state |
					state := handlerStates at: each.
					state ~= 0 and: 
							[newStack := stateStack copyFrom: 1 to: handlerStates size - each + 1.
							newStack add: state.
							self willShift: newStack]]
				ifNone: [nil].
	index isNil] 
			whileTrue: 
				[self dismissErrorToken.
				self isEOFToken 
					ifTrue: 
						[currentToken := startingErrorToken.
						self reportError: anInteger]].
	index - 1 timesRepeat: [self dismissStackTopForErrorRecovery].
	stateStack addLast: (handlerStates at: index).
	nodeStack addLast: startingErrorToken! !

!SmaCCParser methodsFor: 'private' stamp: ' 6/1/07 20:34'!
getNextToken
	currentToken isNil ifTrue: [currentToken := scanner next]! !

!SmaCCParser methodsFor: 'private-error handling' stamp: ' 6/1/07 20:34'!
handleError: anInteger 
	errorToken isNil ifTrue: [errorToken := currentToken].
	(self isEOFToken or: [self hasErrorHandler not]) 
		ifTrue: [self reportError: anInteger].
	self findErrorHandlerIfNoneUseErrorNumber: anInteger! !

!SmaCCParser methodsFor: 'private-error handling' stamp: ' 6/1/07 20:34'!
hasErrorHandler
	^self errorHandlerStates anySatisfy: [:each | each ~~ 0]! !

!SmaCCParser methodsFor: 'initialize-release' stamp: ' 6/1/07 20:34'!
initialize
	nodeStack := OrderedCollection new! !

!SmaCCParser methodsFor: 'testing' stamp: ' 6/1/07 20:34'!
isEOFToken
	^currentToken id first = self emptySymbolTokenId! !

!SmaCCParser methodsFor: 'standard reduction rules' stamp: ' 6/1/07 20:34'!
liftFirstValue: aCollection 
	^aCollection first! !

!SmaCCParser methodsFor: 'standard reduction rules' stamp: ' 6/1/07 20:34'!
liftLastValue: aCollection 
	^aCollection last! !

!SmaCCParser methodsFor: 'standard reduction rules' stamp: ' 6/1/07 20:34'!
liftSecondValue: aCollection 
	^aCollection at: 2! !

!SmaCCParser methodsFor: 'standard reduction rules' stamp: ' 6/1/07 20:34'!
nil
	^nil! !

!SmaCCParser methodsFor: 'accessing' stamp: ' 6/1/07 20:34'!
parse
	self setDefaultStartingStateIfNone.
	self performParsingLoop.
	^nodeStack last! !

!SmaCCParser methodsFor: 'private' stamp: ' 6/1/07 20:34'!
performAction: action 
	| value actionType |
	actionType := action bitAnd: self actionMask.
	value := action bitShift: -2.
	actionType == self shiftAction 
		ifTrue: [self shift: value]
		ifFalse: 
			[actionType == self reduceAction 
				ifTrue: [self reduce: value]
				ifFalse: [self handleError: value]]! !

!SmaCCParser methodsFor: 'private' stamp: ' 6/1/07 20:34'!
performParsingLoop
	| action |
	
	[self getNextToken.
	action := self actionForCurrentToken.
	action = self acceptAction] 
			whileFalse: [self performAction: action].
	self checkForErrors! !

!SmaCCParser methodsFor: 'private' stamp: ' 6/1/07 20:34'!
performReduceMethod: aSymbol with: items 
	^aSymbol last == $: 
		ifTrue: [self perform: aSymbol with: items]
		ifFalse: [self perform: aSymbol]! !

!SmaCCParser methodsFor: 'accessing' stamp: ' 6/1/07 20:34'!
position
	^currentToken isNil 
		ifTrue: [scanner position]
		ifFalse: [currentToken startPosition]! !

!SmaCCParser methodsFor: 'private' stamp: ' 6/1/07 20:34'!
reduce: anInteger 
	| reduceEntry items size |
	reduceEntry := self reduceTable at: anInteger.
	items := OrderedCollection new: (size := reduceEntry at: 2).
	size timesRepeat: 
			[items addFirst: nodeStack removeLast.
			stateStack removeLast].
	nodeStack add: (self performReduceMethod: (reduceEntry at: 3) with: items).
	stateStack add: ((self actionFor: (reduceEntry at: 1)) bitShift: -2)! !

!SmaCCParser methodsFor: 'private' stamp: ' 6/1/07 20:34'!
reduceAction
	^2r10! !

!SmaCCParser methodsFor: 'private' stamp: ' 6/1/07 20:34'!
reduceFor: aCollection 
	| newCollection item |
	(aCollection allSatisfy: [:each | each class ~~ OrderedCollection]) 
		ifTrue: [^aCollection].
	aCollection first class == OrderedCollection 
		ifTrue: 
			[newCollection := aCollection first.
			2 to: aCollection size
				do: 
					[:i | 
					item := aCollection at: i.
					item class = OrderedCollection 
						ifTrue: [newCollection addAll: item]
						ifFalse: [newCollection add: item]].
			^newCollection].
	newCollection := OrderedCollection new.
	aCollection do: 
			[:each | 
			each class == OrderedCollection 
				ifTrue: [newCollection addAll: each]
				ifFalse: [newCollection add: each]].
	^newCollection! !

!SmaCCParser methodsFor: 'accessing' stamp: ' 6/1/07 20:34'!
reduceTable
	^self subclassResponsibility! !

!SmaCCParser methodsFor: 'private-error handling' stamp: ' 6/1/07 20:34'!
reportError: anInteger 
	self reportErrorMessage: (anInteger = 0 
				ifTrue: ['Token not expected']
				ifFalse: [self errorTable at: anInteger])! !

!SmaCCParser methodsFor: 'private-error handling' stamp: 'lr 1/6/2007 22:43'!
reportErrorMessage: aString 
	SmaCCParserError new 
		tag: self;
		signal: aString! !

!SmaCCParser methodsFor: 'accessing' stamp: ' 6/1/07 20:34'!
scanner
	^scanner! !

!SmaCCParser methodsFor: 'initialize-release' stamp: ' 6/1/07 20:34'!
scanner: aScanner 
	scanner := aScanner! !

!SmaCCParser methodsFor: 'private' stamp: ' 6/1/07 20:34'!
setDefaultStartingStateIfNone
	stateStack isNil 
		ifTrue: [self setStartingState: self class defaultStartingState]! !

!SmaCCParser methodsFor: 'initialize-release' stamp: ' 6/1/07 20:34'!
setStartingState: startingState 
	stateStack := OrderedCollection with: startingState! !

!SmaCCParser methodsFor: 'private' stamp: ' 6/1/07 20:34'!
shift: stateIndex 
	stateStack add: stateIndex.
	nodeStack add: currentToken.
	currentToken := nil! !

!SmaCCParser methodsFor: 'private' stamp: ' 6/1/07 20:34'!
shiftAction
	^2r01! !

!SmaCCParser methodsFor: 'standard reduction rules' stamp: ' 6/1/07 20:34'!
stringValue: anOrderedCollection 
	^anOrderedCollection first value! !

!SmaCCParser methodsFor: 'accessing' stamp: ' 6/1/07 20:34'!
transitionTable
	^self subclassResponsibility! !

!SmaCCParser methodsFor: 'private-error handling' stamp: ' 6/1/07 20:34'!
willShift: potentialStateStack 
	| action compoundAction reduceEntry size |
	compoundAction := self actionForState: potentialStateStack last
				and: currentToken id first.
	action := compoundAction bitAnd: self actionMask.
	action == self shiftAction ifTrue: [^true].
	action == self reduceAction 
		ifTrue: 
			[reduceEntry := self reduceTable at: (compoundAction bitShift: -2).
			size := reduceEntry at: 2.
			size timesRepeat: [potentialStateStack removeLast].
			potentialStateStack 
				add: ((self actionForState: potentialStateStack last
						and: (reduceEntry at: 1)) bitShift: -2).
			^self willShift: potentialStateStack].
	^false! !

Object subclass: #SmaCCScanner
	instanceVariableNames: 'stream start matchActions matchEnd currentCharacter outputStream lastOutputStreamMatchPosition lastMatchWasEmpty returnMatchBlock state'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SmaCC-Runtime'!

SmaCCScanner class
	instanceVariableNames: 'keywordMap'!

!SmaCCScanner commentStamp: '<historical>' prior: 0!
SmaCCScanner is an abstract class that represents a scanner for the parser. The scanner converts its string input into SmaCCToken objects that the parser then uses for its parsing.

Subclasses must implement the following messages:
	accessing
		scanForToken

Instance Variables:
	currentCharacter	<Character>	the current character we are scanning
	lastMatchWasEmpty	<Boolean>	was our last scanning match an empty string -- don't allow two empty matches in a row
	lastOutputStreamMatchPosition	<Integer>	the position in the outputStream of the last match
	matchActions	<Array | Symbol>	the actions for the last match (a symbol means that the action should be performed on the scanner)
	matchEnd	<Integer>	the position of the last match in the stream (our input stream)
	outputStream	<PositionableStream>	the matched characters go in this stream. After a match is made, we take this stream's contents and create a token object.
	returnMatchBlock	<BlockClosure>	when we match a token evaluate this block with the token (hack to return from multiple levels)
	start	<Integer>	the starting position of a match in the stream
	stream	<Stream>	our input

!

SmaCCScanner class
	instanceVariableNames: 'keywordMap'!

!SmaCCScanner class methodsFor: 'accessing' stamp: ' 6/1/07 21:19'!
frequencyTable
	^#(1)! !

!SmaCCScanner class methodsFor: 'class initialization' stamp: ' 6/1/07 20:34'!
initialize
	self initializeKeywordMap! !

!SmaCCScanner class methodsFor: 'class initialization' stamp: ' 6/1/07 20:34'!
initializeKeywordMap
	keywordMap := Dictionary new! !

!SmaCCScanner class methodsFor: 'accessing' stamp: ' 6/1/07 20:34'!
keywordMap
	keywordMap isNil ifTrue: [self initializeKeywordMap].
	^keywordMap! !

!SmaCCScanner class methodsFor: 'testing' stamp: ' 6/1/07 20:34'!
needsLineNumbers
	"Redefine to return true, if you need line number information"

	^false! !

!SmaCCScanner class methodsFor: 'instance creation' stamp: ' 6/1/07 20:34'!
on: aStream 
	^(self new)
		on: (self needsLineNumbers 
					ifTrue: [SmaCCLineNumberStream on: aStream]
					ifFalse: [aStream]);
		yourself! !

!SmaCCScanner methodsFor: 'testing' stamp: ' 6/1/07 20:34'!
atEnd
	^stream atEnd! !

!SmaCCScanner methodsFor: 'private' stamp: ' 6/1/07 20:34'!
checkForKeyword: aString 
	| stateMap action |
	action := matchActions isSymbol 
				ifTrue: [matchActions]
				ifFalse: [matchActions first].
	stateMap := self class keywordMap at: action ifAbsent: [nil].
	stateMap isNil ifTrue: [^self].
	matchActions := stateMap at: (self keywordFor: aString)
				ifAbsent: [matchActions].
	matchActions isInteger 
		ifTrue: [matchActions := Array with: matchActions with: action]! !

!SmaCCScanner methodsFor: 'private' stamp: ' 6/1/07 20:34'!
checkForValidMatch
	matchActions isNil ifTrue: [self scannerError]! !

!SmaCCScanner methodsFor: 'default token handling' stamp: ' 6/1/07 20:34'!
comment
	"In case someone wants to record the comments"

	^self whitespace! !

!SmaCCScanner methodsFor: 'accessing' stamp: ' 6/1/07 20:34'!
contents
	| writeStream token |
	writeStream := WriteStream with: Array new.
	[self atEnd] whileFalse: 
			[token := self next.
			token notNil ifTrue: [writeStream nextPut: token]].
	^writeStream contents! !

!SmaCCScanner methodsFor: 'private' stamp: ' 6/1/07 20:34'!
createTokenFor: string 
	| token |
	token := self tokenClass 
				value: string
				start: start + 1
				id: matchActions.
	outputStream reset.
	^token! !

!SmaCCScanner methodsFor: 'accessing' stamp: ' 6/1/07 20:34'!
emptySymbolTokenId
	^self subclassResponsibility! !

!SmaCCScanner methodsFor: 'private' stamp: ' 6/1/07 20:34'!
eofTokenId
	^Array with: self emptySymbolTokenId! !

!SmaCCScanner methodsFor: 'accessing' stamp: ' 6/1/07 20:34'!
errorTokenId
	^self subclassResponsibility! !

!SmaCCScanner methodsFor: 'initialize-release' stamp: ' 6/1/07 20:34'!
initialize
	outputStream := WriteStream with: String new.
	lastMatchWasEmpty := true.
	state := #default! !

!SmaCCScanner methodsFor: 'private-utility' stamp: ' 6/1/07 20:34'!
keywordFor: aString 
	"Subclasses can override this to ignore case"

	^aString! !

!SmaCCScanner methodsFor: 'accessing' stamp: ' 6/1/07 20:34'!
lineNumber
	"This requires the stream to be a line number stream (see the #needsLineNumbers class method)."

	^stream lineNumber! !

!SmaCCScanner methodsFor: 'accessing' stamp: ' 6/1/07 20:34'!
next
	self resetScanner.
	returnMatchBlock := [:match | ^match].
	^self scanForToken! !

!SmaCCScanner methodsFor: 'initialize-release' stamp: ' 6/1/07 20:34'!
on: aStream 
	stream := aStream.
	start := stream position! !

!SmaCCScanner methodsFor: 'accessing' stamp: ' 6/1/07 20:34'!
position
	^stream position! !

!SmaCCScanner methodsFor: 'accessing' stamp: ' 6/1/07 20:34'!
position: anInteger
	^stream position: anInteger! !

!SmaCCScanner methodsFor: 'private' stamp: ' 6/1/07 20:34'!
recordAndReportMatch: aCollection 
	^self
		recordMatch: aCollection;
		reportLastMatch! !

!SmaCCScanner methodsFor: 'private' stamp: ' 6/1/07 20:34'!
recordMatch: aCollection 
	matchActions := aCollection.
	matchEnd := stream position.
	lastOutputStreamMatchPosition := outputStream position! !

!SmaCCScanner methodsFor: 'private' stamp: ' 6/1/07 20:34'!
reportLastMatch
	"The scanner has found the end of a token and must report it"

	| string token |
	self checkForValidMatch.
	self resetOutputToLastMatch.
	stream position: matchEnd.
	string := outputStream contents.
	self checkForKeyword: string.
	token := matchActions isSymbol 
				ifTrue: [self perform: matchActions]
				ifFalse: [self createTokenFor: string].
	matchActions := nil.
	^token! !

!SmaCCScanner methodsFor: 'private' stamp: ' 6/1/07 20:34'!
resetOutputToLastMatch
	outputStream position: lastOutputStreamMatchPosition.
	lastOutputStreamMatchPosition == 0 
		ifTrue: 
			[lastMatchWasEmpty ifTrue: [self scannerError].
			lastMatchWasEmpty := true]
		ifFalse: [lastMatchWasEmpty := false]! !

!SmaCCScanner methodsFor: 'private' stamp: 'lr 1/7/2007 02:23'!
resetScanner
	start := stream position.
	outputStream reset.
	lastOutputStreamMatchPosition := 0! !

!SmaCCScanner methodsFor: 'accessing' stamp: ' 6/1/07 20:34'!
scanForToken
	^self subclassResponsibility! !

!SmaCCScanner methodsFor: 'private' stamp: ' 6/1/07 20:34'!
scannerError
	| string |
	(stream atEnd and: [start == stream position]) 
		ifTrue: 
			[string := ''.
			matchActions := self eofTokenId]
		ifFalse: 
			[stream position: start.
			string := String with: stream next.
			matchActions := #(0)].
	returnMatchBlock value: (self createTokenFor: string)! !

!SmaCCScanner methodsFor: 'accessing' stamp: ' 6/1/07 20:34'!
state
	^state! !

!SmaCCScanner methodsFor: 'accessing' stamp: ' 6/1/07 20:34'!
state: aSymbol
	state := aSymbol! !

!SmaCCScanner methodsFor: 'private' stamp: ' 6/1/07 20:34'!
step
	stream atEnd ifTrue: [^returnMatchBlock value: self reportLastMatch].
	currentCharacter := stream next.
	outputStream nextPut: currentCharacter! !

!SmaCCScanner methodsFor: 'private' stamp: ' 6/1/07 20:34'!
tokenClass
	^SmaCCToken! !

!SmaCCScanner methodsFor: 'default token handling' stamp: ' 6/1/07 20:34'!
whitespace
	"By default, eat the whitespace"

	self resetScanner.
	^self scanForToken! !

Object subclass: #SmaCCToken
	instanceVariableNames: 'start id value'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SmaCC-Runtime'!

!SmaCCToken commentStamp: '<historical>' prior: 0!
SmaCCTokens are used as the interface objects between scanner and parser. They hold the string that was scanned and its position information. Also, included in the token is its id. The id specifies what type of token it is.

Instance Variables:
	id	<Array of: Integer>	the list of possible token types this represents. There can be overlapping tokens, so we list all of the id here. The default parser only looks at the first id, but we can redefine this behavior in a subclass to look at all possibilities until we find a valid token.
	start	<Integer>	the starting position of the token in the original input
	value	<Object>	the value of our token (normally a string, but could be anything)
!

!SmaCCToken class methodsFor: 'instance creation' stamp: ' 6/1/07 20:34'!
value: aString start: anInteger id: anObject 
	^(self new)
		value: aString
			start: anInteger
			id: anObject;
		yourself! !

!SmaCCToken methodsFor: 'accessing' stamp: ' 6/1/07 20:34'!
id
	^id! !

!SmaCCToken methodsFor: 'printing' stamp: ' 6/1/07 20:34'!
printOn: aStream 
	aStream
		nextPut: ${;
		nextPutAll: self value;
		nextPut: $(;
		nextPutAll: self startPosition printString;
		nextPut: $,;
		nextPutAll: self stopPosition printString;
		nextPut: $,;
		nextPutAll: self id printString;
		nextPutAll: ')}'! !

!SmaCCToken methodsFor: 'accessing' stamp: ' 6/1/07 20:34'!
startPosition
	^start! !

!SmaCCToken methodsFor: 'accessing' stamp: ' 6/1/07 20:34'!
stopPosition
	^start + value size - 1! !

!SmaCCToken methodsFor: 'accessing' stamp: ' 6/1/07 20:34'!
value
	^value! !

!SmaCCToken methodsFor: 'initialize-release' stamp: ' 6/1/07 20:34'!
value: anObject start: startPositionInteger id: anInteger 
	value := anObject.
	start := startPositionInteger.
	id := anInteger! !

Stream subclass: #SmaCCLineNumberStream
	instanceVariableNames: 'sourceStream previousWasCR eolPositions lastPosition'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SmaCC-Runtime'!

!SmaCCLineNumberStream commentStamp: '<historical>' prior: 0!
SmaCCLineNumberStream is a wrapper for streams that calculates line numbers.

Instance Variables:
	eolPositions	<OrderedCollection of: Integer>	the positions of each end of line
	lastPosition	<Integer>	the position of the last character that we have calculated the end of line information for (we know the line number for all characters before this position and don't know anything about the characters after this position)
	previousWasCR	<Boolean>	was the previous character a CR. This is used for CR LF streams. A CR LF combination should only increment the line counter by 1
	sourceStream	<Stream>	the stream that we are wrapping

!

!SmaCCLineNumberStream class methodsFor: 'instance creation' stamp: ' 6/1/07 20:34'!
on: aReadStream 
	^(self basicNew)
		on: aReadStream;
		yourself! !

!SmaCCLineNumberStream methodsFor: 'testing' stamp: ' 6/1/07 20:34'!
atEnd
	^sourceStream atEnd! !

!SmaCCLineNumberStream methodsFor: 'accessing' stamp: ' 6/1/07 20:34'!
columnNumber
	^self columnNumberFor: self position! !

!SmaCCLineNumberStream methodsFor: 'accessing' stamp: ' 6/1/07 20:34'!
columnNumberFor: anInteger 
	^anInteger - (eolPositions at: (self lineNumberFor: anInteger)) + 1! !

!SmaCCLineNumberStream methodsFor: 'accessing' stamp: ' 6/1/07 20:34'!
contents
	^sourceStream contents! !

!SmaCCLineNumberStream methodsFor: 'error handling' stamp: ' 6/1/07 20:34'!
doesNotUnderstand: aMessage 
	^sourceStream perform: aMessage selector withArguments: aMessage arguments! !

!SmaCCLineNumberStream methodsFor: 'accessing' stamp: ' 6/1/07 20:34'!
flush
	^sourceStream flush! !

!SmaCCLineNumberStream methodsFor: 'testing' stamp: ' 6/1/07 20:34'!
isReadable
	^sourceStream isReadable! !

!SmaCCLineNumberStream methodsFor: 'testing' stamp: ' 6/1/07 20:34'!
isWritable
	^sourceStream isWritable! !

!SmaCCLineNumberStream methodsFor: 'accessing' stamp: ' 6/1/07 20:34'!
lineNumber
	^self lineNumberFor: self position! !

!SmaCCLineNumberStream methodsFor: 'accessing' stamp: ' 6/1/07 20:34'!
lineNumberFor: anInteger
	| index start stop pos |
	pos := anInteger.
	pos >= eolPositions last ifTrue: [^eolPositions size].
	start := 1.
	stop := eolPositions size.
	[start + 1 < stop] whileTrue: 
			[index := (start + stop) // 2.
			(eolPositions at: index) <= pos 
				ifTrue: [start := index]
				ifFalse: [stop := index]].
	^start! !

!SmaCCLineNumberStream methodsFor: 'accessing' stamp: ' 6/1/07 20:34'!
next
	| character streamPosition |
	character := sourceStream next.
	streamPosition := sourceStream position.
	streamPosition - 1 >= lastPosition 
		ifTrue: 
			[lastPosition := streamPosition.
			character == Character cr 
				ifTrue: 
					[eolPositions add: lastPosition.
					previousWasCR := true]
				ifFalse: 
					[character == Character lf 
						ifTrue: 
							[previousWasCR 
								ifTrue: [eolPositions at: eolPositions size put: lastPosition]
								ifFalse: [eolPositions add: lastPosition]].
					previousWasCR := false]].
	^character! !

!SmaCCLineNumberStream methodsFor: 'accessing' stamp: ' 6/1/07 20:34'!
nextPut: anObject 
	^sourceStream nextPut: anObject! !

!SmaCCLineNumberStream methodsFor: 'initialize-release' stamp: ' 6/1/07 20:34'!
on: aReadStream 
	sourceStream := aReadStream.
	eolPositions := OrderedCollection with: aReadStream position.
	lastPosition := aReadStream position.
	previousWasCR := false! !

!SmaCCLineNumberStream methodsFor: 'accessing' stamp: ' 6/1/07 20:34'!
peek
	^sourceStream peek! !

!SmaCCLineNumberStream methodsFor: 'accessing' stamp: ' 6/1/07 20:34'!
position
	^sourceStream position! !

!SmaCCLineNumberStream methodsFor: 'accessing' stamp: ' 6/1/07 20:34'!
position: anInteger 
	anInteger > lastPosition 
		ifTrue: 
			[sourceStream position: lastPosition.
			[sourceStream position < anInteger and: [sourceStream atEnd not]] 
				whileTrue: [self next]]
		ifFalse: [sourceStream position: anInteger]! !

!SmaCCLineNumberStream methodsFor: 'accessing' stamp: ' 6/1/07 20:34'!
skip: anInteger
	^self position: self position + anInteger! !

!SmaCCLineNumberStream methodsFor: 'accessing' stamp: 'lr 1/6/2007 20:56'!
upTo: aCharacter 
	| stream char |
	stream := WriteStream on: String new.
	[self atEnd or: [ (char := self next) = aCharacter]] 
		whileFalse: [stream nextPut: char].
	^ stream contents! !
SmaCCScanner initialize!
